from contextlib import contextmanager
from csv_plot.csv.splitted_gettable import SplittedGettable
from pathlib import Path
from typing import IO, Iterator, List, Optional, Tuple, Union

from .gettable import Gettable


class OffsetError(Exception):
    pass


class TextFileNotPaddedError(Exception):
    pass


class PaddedTextFile(Gettable):
    """Represent a padded text file, where lines are reachable with O(1) complexity.

    A padded text file is a text file where all lines have exactly the same length.
    In general, lines are right padded with white spaces.
    The last line MUST also contain a carriage return.

    Only line(s) you request will be loaded in memory.

    Usage:
    padded_text_file = PaddedTextFile(<file_descriptor>, <file_zise>, <offset>)

    # Get the number of lines
    len(padded_text_file)

    # Get the third line of the file
    padded_text_file[2]

    # Get the last line of the file
    padded_text_file[-1]

    # Get an iterator on lines between the third line (included) and the last line
    # (excluded)
    padded_text_file.get(start=2, stop=-1)

    # Get all lines between the third line (included) and the last line (excluded)
    # Warning: All lines in the selected range will be loaded into memory.
    #          For example: padded_text_file[:] will load all the file in memory.
    #          If possible, use padded_text_file.get(start=a, stop=b) instead of
    #                           padded_text_file[a, b]
    padded_text_file[2:-1]

    # Get the size (number of characters) of one line
    padded_text_file.line_size
    """

    def __init__(self, file_descriptor: IO, file_size: int, offset: int) -> None:
        """Constructor.

        file_descriptor: The file descriptor pointing to the padded CSV file
        file_size      : The file size (in bytes) of the padded CSV file pointed by
                         `file_descriptor`
        offset         : The number of first line(s) to skip. Must be >= 0

        If not 0 <= `offset` <= number of lines, an `OffsetError` is raised.

        If at least one line of the file pointed by `file_descriptor` has not the same
        length than others, a `TextFileNotPaddedError` is raised.
        """
        file_descriptor.seek(0)
        self.__offset = offset
        self.__file_descriptor = file_descriptor

        first_line = next(file_descriptor)

        self.line_size = len(first_line)
        self.__len = file_size // self.line_size

        if not 0 <= offset <= self.__len:
            raise OffsetError(f"Offset must be in [0;{self.__len}]")

        if file_size % self.line_size != 0:
            raise TextFileNotPaddedError("text file is not padded")

    def __get_start_and_stop(
        self,
        slice_start: Optional[int],
        slice_stop: Optional[int],
    ) -> Tuple[int, int]:
        """Get `start` and `stop`."""
        start = (
            (
                self.__offset + slice_start
                if slice_start >= 0
                else max(self.__offset, self.__len + slice_start)
            )
            if slice_start is not None
            else self.__offset
        )

        stop = (
            (
                (
                    min(slice_stop + self.__offset, self.__len)
                    if slice_stop >= 0
                    else self.__len + slice_stop
                )
            )
            if slice_stop is not None
            else self.__len
        )

        return start, stop

    def __len__(self):
        """Return the number of lines of the file."""
        return self.__len - self.__offset

    def __getitem__(
        self, line_number_or_slice: Union[int, slice]
    ) -> Union[str, List[str]]:
        """Get a given line or a given slice of lines.

        line_number_or_slice: The line number or the slice where lines will be retrieved
        """

        def handle_line_number(line_number: int) -> str:
            real_line_number = (
                line_number + self.__offset
                if line_number >= 0
                else self.__len + line_number
            )

            if not self.__offset <= real_line_number < self.__len:
                raise IndexError("list index out of range")

            self.__move_cursor(real_line_number)
            return next(self.__file_descriptor).rstrip()

        def handle_slice(slice: slice) -> List[str]:
            start, stop = self.__get_start_and_stop(slice.start, slice.stop)

            self.__move_cursor(start)
            return [next(self.__file_descriptor).rstrip() for _ in range(start, stop)]

        if isinstance(line_number_or_slice, int):
            return handle_line_number(line_number_or_slice)
        elif isinstance(line_number_or_slice, slice):
            return handle_slice(line_number_or_slice)

    def get(
        self, start: Optional[int] = None, stop: Optional[int] = None
    ) -> Iterator[str]:
        """Return an iterator on a given slice of lines.

        start: The first line of slice (included)
        stop : The last line of slice (excluded)
        """
        real_start, real_stop = self.__get_start_and_stop(start, stop)
        self.__move_cursor(real_start)

        for _ in range(real_start, real_stop):
            yield next(self.__file_descriptor).rstrip()

    def __move_cursor(self, line_number: int) -> None:
        """Move cursor of file descriptor to `line_number`.

        line_number: The line number where to move the cursor
        """
        self.__file_descriptor.seek(self.line_size * line_number, 0)


class SplittedPaddedTextFile(SplittedGettable):
    """Splitted Padded Text File"""

    def __init__(self, files_descriptor_and_size: List[Tuple[IO, int]], offset: int):
        """Initializer.

        files_descriptor_and_size:
            A list of tuple (<file descriptor>, <size of corresponding file>)

        offset: The offset to apply to the first file descriptor
        """
        padded_text_files = [
            PaddedTextFile(file_descriptor, file_size, offset=0)
            for file_descriptor, file_size in files_descriptor_and_size
        ]

        super().__init__(padded_text_files, offset)  # type:ignore


@contextmanager
def padded_text_file(path: Path, offset: int = 0) -> Iterator[PaddedTextFile]:
    """Represent a padded text file, where lines are reachable with O(1) complexity.

    A padded text file is a text file where all lines have exactly the same length.
    In general, lines are right padded with white spaces.
    The last line MUST also contain a carriage return.

    Only line(s) you request will be loaded in memory.

    Usage:
    with padded_text_file(<file_path>, <offset>) as ptf:
        # Get the number of lines
        len(ptf)

        # Get the third line of the file
        ptf[2]

        # Get the last line of the file
        ptf[-1]

        # Get an iterator on lines between the third line (included) and the last line
        # (excluded)
        ptf.get(start=2, stop=-1)

        # Get all lines between the third line (included) and the last line (excluded)
        Warning: All lines in the selected range will be loaded into memory.
                For example: ptf[:] will load all the file in memory.
                If possible, use ptf.get(start=a, stop=b) instead of ptf[a, b]
        ptf[2:-1]
    """
    with path.open() as file_descriptor:
        yield PaddedTextFile(file_descriptor, path.stat().st_size, offset)



from bisect import bisect_left, bisect_right
from contextlib import ExitStack, contextmanager
from enum import Enum
from pathlib import Path
from typing import IO, Any, Dict, Iterator, List, Optional, Set, Tuple, Union

from .gettable import Gettable
from .padded_csv_file import _PaddedCSVFile


class Side(Enum):
    Left = "left"
    Right = "right"


class _SortedPaddedCSVFile(Gettable):
    """Represent a padded CSV file with one sorted column, where all lines are reachable
    through the sorted column with O(log(n)) complexity.

    A padded CSV file is a CSV file where all lines have exactly the same length.
    In general, lines are right padded with white spaces.
    The last line MUST also contain a carriage return.

    Only line(s) you request will be load in memory.

    Usage:
    sorted_padded_csv_file = _SortedPaddedCSVFile(<files_descriptors_and_size>,
                                                  <x_column_and_type_tuple>,
                                                  <y_column_and_type_tuples>
                                                 )

    Example: With the following file represented by <file_descriptor>:
    a,b,c,d
    1,2,3,4
    5,6,7,8
    9,10,11,12
    13,14,15,16
    17,18,19,20

    Here all columns are sorted, but in this example, only the column "c" has to be.
    sorted_padded_csv_file = _SortedPaddedCSVFile(<files_descriptors_and_size>,
                                                  ("c", int)
                                                  [("d", int), ("b", int)]
                                                 )

    # Get the number of lines
    len(sorted_padded_csv_file) # = 5

    # Get the line corresponding to c == 7
    sorted_padded_csv_file[7] # = (3, [8, 6])

    # Get an iterator on all lines corresponding to 6.5 <= c <= 15.5
    sorted_padded_csv_file.get(start=6.5, stop=15.5)

    # Get all lines corresponding to 6.5 <= c <= 15.5
    sorted_padded_csv_file[6.5:15.5] # = [(7, [8, 6]), (11, [12, 10]), (15, [16, 14])]

    # Warning: All lines in the selected range will be loaded into memory.
    #          For example: sorted_padded_csv_file[:] will load all the file in memory.
    #          If possible, use sorted_padded_csv_file.get(start=a, stop=b) instead of
    #                           sorted_padded_csv_file[a, b]
    """

    def __init__(
        self,
        files_descriptors_and_size: List[Tuple[IO, IO, int]],
        x_and_type: Tuple[str, type],
        ys: List[str],
    ) -> None:
        """Constructor.

        files_descriptors_and_size:
            A list of tuple, where each tuple contains
            - A first file descriptor to the file
            - A second (and distinct) file descriptor to the file
            - The size (in bytes) of the file

        x_and_type:
            A tuple containing
            - The name of the column on which requests will be done
            - The type of the column on which requests will be done

        ys_and_types:
            A list of tuples where each tuple has:
            - The name of the column
            - The type of the column

        If at least one line of the file pointed by `file_descriptor` has not the same
        length than others, a `TextFileNotPaddedError` is raised.
        """
        files_descriptor_1_and_size = [
            (file_descriptor, size)
            for file_descriptor, _, size in files_descriptors_and_size
        ]

        files_descriptor_2_and_size = [
            (file_descriptor, size)
            for _, file_descriptor, size in files_descriptors_and_size
        ]

        self.__x_file = _PaddedCSVFile(
            files_descriptor_1_and_size, [x_and_type], unwrap_if_one_column=True
        )

        self.__ys_file = _PaddedCSVFile(
            files_descriptor_2_and_size, [(y, float) for y in ys]
        )

    def __get_line_number_of(self, x: Any, side: Side) -> int:
        return {Side.Left: bisect_left, Side.Right: bisect_right}[side](
            self.__x_file, x  # type: ignore
        )

    def __get_start_stop(self, slice: slice) -> Tuple[Optional[int], Optional[int]]:
        start = (
            self.__get_line_number_of(slice.start, Side.Left)
            if slice.start is not None
            else None
        )

        stop = (
            self.__get_line_number_of(slice.stop, Side.Right)
            if slice.stop is not None
            else None
        )

        return start, stop

    def __len__(self):
        """Return the number of lines of the file (excluding the header)."""
        return len(self.__x_file)

    def __getitem__(
        self, x_or_slice: Union[Any, slice]
    ) -> Union[Tuple[Any, List], List[Tuple[Any, List]]]:
        """Get the line where `x` equals `x_or_slice`, or get all the lines where
        `slice.start <= x <= slice.stop`.

        line_number_or_slice: The x value(s) corresponding to the lines to retrieve
        """

        def handle_x(x: Any) -> Tuple[Any, List]:
            line_number = self.__get_line_number_of(x, Side.Left)
            closest_x = self.__x_file[line_number]

            if closest_x != x:
                raise IndexError("not found")

            ys = self.__ys_file[line_number]
            return (x, ys)

        def handle_slice(slice: slice) -> List[Tuple[Any, List]]:
            start, stop = self.__get_start_stop(slice)

            xs = self.__x_file[start:stop]
            yss = self.__ys_file[start:stop]

            return list(zip(xs, yss))

        if isinstance(x_or_slice, slice):
            return handle_slice(x_or_slice)
        else:
            return handle_x(x_or_slice)

    def get(
        self, start: Optional[Any] = None, stop: Optional[Any] = None
    ) -> Iterator[Tuple[Any, List]]:
        """Return an iterator on a given slice of lines.

        start: The value of `x` corresponding to the first line
        stop : The value of `x` corresponding to the last line
        """
        start_, stop_ = self.__get_start_stop(slice(start, stop))

        x_iterator = self.__x_file.get(start_, stop_)
        ys_iterator = self.__ys_file.get(start_, stop_)

        for x, ys in zip(x_iterator, ys_iterator):
            yield x, ys

    def number_of_lines_between(
        self, start: Optional[Any] = None, stop: Optional[Any] = None
    ) -> int:
        """Get the number of lines between `start` and `stop`.

        start: The value of `x` corresponding to the first line
        stop : The value of `x` corresponding to the last line
        """
        start_, stop_ = self.__get_start_stop(slice(start, stop))
        return (stop_ if stop_ is not None else len(self)) - (
            start_ if start_ is not None else 0
        )


@contextmanager
def sorted_padded_csv_file(
    path: Path,
    x_and_type: Tuple[str, type],
    ys: List[str],
) -> Iterator[_SortedPaddedCSVFile]:
    """Represent a padded CSV file with one sorted column, where all lines are reachable
    through the sorted column with O(log(n)) complexity.

    A padded CSV file is a CSV file where all lines have exactly the same length.
    In general, lines are right padded with white spaces.
    The last line MUST also contain a carriage return.

    Only line(s) you request will be load in memory.

    Usage:
    with sorted_padded_csv_file as spcf:
        ...

    Example: With the following file represented by <file_descriptor>:
    a,b,c,d
    1,2,3,4
    5,6,7,8
    9,10,11,12
    13,14,15,16
    17,18,19,20

    Here all columns are sorted, but in this example, only the column "c" has to be.

    with sorted_padded_csv_file(<file_path>, ("c", int), [("d", int), ("b", int)]) as spcf:
        # Get the number of lines
        len(spcf) # = 5

        # Get the line corresponding to c == 7
        spcf[7] # = (3, [8, 6])

        # Get an iterator on all lines corresponding to 6.5 <= c <= 15.5
        spcf.get(start=6.5, stop=15.5)

        # Get all lines corresponding to 6.5 <= c <= 15.5
        spcf[6.5:15.5] # = [(7, [8, 6]), (11, [12, 10]), (15, [16, 14])]

        # <file_path> could be either a file, or a directory.
        # If it is a directory, each file inside should follow the convention <int>.csv.
        # The file which has the smaller <int> should have the header as a first line.

        # Warning: All lines in the selected range will be loaded into memory.
        #          For example: spcf[:] will load all the file in memory.
        #          If possible, use spcf.get(start=a, stop=b) instead of
        #                           spcf[a, b]
    """
    if path.is_file():
        with path.open() as file_descriptor_1, path.open() as file_descriptor_2:
            yield _SortedPaddedCSVFile(
                [(file_descriptor_1, file_descriptor_2, path.stat().st_size)],
                x_and_type,
                ys,
            )
    else:
        paths = sorted(path.glob("*.csv"), key=lambda item: int(item.stem))

        with ExitStack() as stack:
            files_descriptors_and_size = [
                (
                    stack.enter_context(path.open()),
                    stack.enter_context(path.open()),
                    path.stat().st_size,
                )
                for path in paths
            ]

            yield _SortedPaddedCSVFile(files_descriptors_and_size, x_and_type, ys)  # type: ignore



from itertools import accumulate
from typing import Any, Iterator, List, Optional, Tuple, Union

from .gettable import Gettable


class SplittedGettable:
    """A SplittedGettable represents a merge of several `Gettable` objects.

    Example with the following custom `GettableList`:

    class GettableList(list):
        def get(self, start: Optional[int], stop: Optional[int]) -> Iterable:
            for index in range(
                start if start is not None else 0,
                stop if stop is not None else len(self),
            ):
                yield self[index]

    one = GettableList(["a", "b", "c"])
    two = GettableList(["d"])
    three = GettableList(["e", "f", "g"])

    splitted_gettable = SplittedGettable([one, two, three])

    splitted_gettable[1] == "b"
    splitted_gettable[-2] == "f"
    splitted_gettable[3:-2] == ["d", "e", "f"]

    # Warning: Getting a slice will load all retrieved values into memory.
    #          To get an iterator (which load only what is needed) do:

    splitted_gettable.get(3, -2)
    """

    def __init__(self, gettables: List[Gettable], offset: int) -> None:
        """Initializer.

        gettables: List of Gettable objects
        offset: The offset (apply only on the first Gettable of the list)
        """
        gettables_lengths = [len(gettable) for gettable in gettables]

        self.__gettables = gettables
        self.__len = sum(gettables_lengths)
        self.__offset = offset

        cumulated_lengths = list(accumulate(gettables_lengths))
        shifted_cumulated_lengths = [0] + cumulated_lengths

        self.__blocks_start_stop = [
            (start, stop)
            for start, stop in zip(
                shifted_cumulated_lengths[:-1], shifted_cumulated_lengths[1:]
            )
        ]

    def __len__(self) -> int:
        """Return total number of elements of all Gettable (excluding the offset)"""
        return self.__len - self.__offset

    def __getitem__(self, index_or_slice: Union[int, slice]) -> Union[Any, List]:
        """Get given value or a given slice of values.

        index_or_slice: The index or the slice where values will be
                        retrieved
        """
        if isinstance(index_or_slice, int):
            block_index, item_index = self.__get_index(index_or_slice)
            return self.__gettables[block_index][item_index]
        elif isinstance(index_or_slice, slice):
            return list(self.get(index_or_slice.start, index_or_slice.stop))

    def get(self, start: Optional[int] = None, stop: Optional[int] = None) -> Iterator:
        slices = self.__get_slices(start, stop)

        iterable_slice_list = [
            (iterable, slice_)
            for iterable, slice_ in zip(self.__gettables, slices)
            if slice_ is not None
        ]

        for iterable, slice_ in iterable_slice_list:
            start, stop = slice_
            yield from iterable.get(start, stop)

    def __get_index(self, index: int) -> Tuple[int, int]:
        real_index = index + self.__offset if index >= 0 else self.__len + index

        if not self.__offset <= real_index < self.__len:
            raise IndexError("list index out of range")

        for block_index, (start, stop) in enumerate(self.__blocks_start_stop):
            if start <= real_index < stop:
                return block_index, real_index - start

        raise RuntimeError("This code should never be reached")  # pragma: no cover

    def __get_slices(
        self, slice_start: Optional[int], slice_stop: Optional[int]
    ) -> List[Optional[Tuple[Optional[int], Optional[int]]]]:
        def compute_item(
            block_start_stop: Tuple[int, int],
            slice_start_stop: Tuple[int, int],
        ) -> Optional[Tuple[Optional[int], Optional[int]]]:
            block_start, block_stop = block_start_stop
            slice_start, slice_stop = slice_start_stop

            if slice_start == slice_stop:
                return None
            elif block_start < block_stop <= slice_start < slice_stop:
                #            [ SLICE [
                # [ BLOCK [
                # ---------------------------------
                #            [ SLICE [
                #    [ BLOCK [
                return None
            elif block_start < slice_start < block_stop < slice_stop:
                #            [ SLICE [
                #        [ BLOCK [
                return slice_start - block_start, None
            elif slice_start < block_start < slice_stop < block_stop:
                #            [ SLICE [
                #                [ BLOCK [
                return None, slice_stop - block_start
            elif slice_start < slice_stop <= block_start < block_stop:
                #            [ SLICE [
                #                    [ BLOCK [
                # ---------------------------------
                #            [ SLICE [
                #                       [ BLOCK [
                return None
            elif slice_start <= block_start < block_stop <= slice_stop:
                #         [    SLICE    [
                #            [ BLOCK [
                # ---------------------------------
                #         [    SLICE [
                #            [ BLOCK [
                # ---------------------------------
                #            [ SLICE    [
                #            [ BLOCK [
                # ---------------------------------
                #            [ SLICE [
                #            [ BLOCK [

                return None, None
            elif block_start <= slice_start < slice_stop <= block_stop:
                #            [ SLICE [
                #         [    BLOCK    [
                # ---------------------------------
                #            [ SLICE [
                #         [    BLOCK [
                # ---------------------------------
                #            [ SLICE [
                #            [ BLOCK    [
                # ---------------------------------
                #            [ SLICE [
                #            [ BLOCK [
                return (
                    slice_start - block_start
                    if slice_start - block_start > 0
                    else None,
                    slice_stop - block_start if slice_stop - block_start > 0 else None,
                )

            raise RuntimeError("This code should never be reached")  # pragma: no cover

        real_slice_start = (
            (
                slice_start + self.__offset
                if slice_start >= 0
                else max(self.__offset, self.__len + slice_start)
            )
            if slice_start is not None
            else self.__offset
        )

        real_slice_stop = (
            (
                (
                    min(slice_stop + self.__offset, self.__len)
                    if slice_stop >= 0
                    else self.__len + slice_stop
                )
            )
            if slice_stop is not None
            else self.__len
        )

        real_slice_start_stop = real_slice_start, real_slice_stop

        return [
            compute_item(block_start_stop, real_slice_start_stop)
            for block_start_stop in self.__blocks_start_stop
        ]



from pathlib import Path
from typing import IO, List, Tuple

import pytest
from pytest import fixture

from ..padded_text_file import (
    OffsetError,
    SplittedPaddedTextFile,
    TextFileNotPaddedError,
    PaddedTextFile,
    padded_text_file,
)
from ..tests import assets


@fixture
def not_padded_file_path() -> Path:
    return Path(assets.__file__).parent / "not_padded.csv"


@fixture
def not_padded_file_descriptor(not_padded_file_path: Path) -> IO:
    return not_padded_file_path.open()


@fixture
def not_padded_file_size(not_padded_file_path: Path) -> int:
    return not_padded_file_path.stat().st_size


@fixture
def padded_file_path() -> Path:
    return Path(assets.__file__).parent / "padded.csv"


@fixture
def splitted_padded_dir_path() -> Path:
    return Path(assets.__file__).parent / "splitted_padded_csv"


@fixture
def padded_file_descriptor(padded_file_path: Path) -> IO:
    return padded_file_path.open()


@fixture
def padded_file_size(padded_file_path: Path) -> int:
    return padded_file_path.stat().st_size


@fixture
def splitted_padded_csv_files_descriptor_size(
    splitted_padded_dir_path: Path,
) -> List[Tuple[IO, int]]:
    paths = [(splitted_padded_dir_path / f"{index}.csv") for index in range(3)]
    return [(path.open(), path.stat().st_size) for path in paths]


def test_not_padded(not_padded_file_descriptor: IO, not_padded_file_size: int) -> None:
    with pytest.raises(TextFileNotPaddedError):
        PaddedTextFile(not_padded_file_descriptor, not_padded_file_size, 0)


def test_len(padded_file_descriptor: IO, padded_file_size: int) -> None:
    padded_text_file = PaddedTextFile(padded_file_descriptor, padded_file_size, 0)
    assert len(padded_text_file) == 6


def test_line(padded_file_descriptor: IO, padded_file_size: int) -> None:
    padded_text_file = PaddedTextFile(padded_file_descriptor, padded_file_size, 0)

    assert padded_text_file[2] == "5,6,7,8"
    assert padded_text_file[-1] == "17,18,19,20"

    with pytest.raises(IndexError):
        padded_text_file[6]

    with pytest.raises(IndexError):
        padded_text_file[-7]


def test_slice(padded_file_descriptor: IO, padded_file_size: int) -> None:
    padded_text_file = PaddedTextFile(padded_file_descriptor, padded_file_size, 0)

    assert (
        padded_text_file[:]
        == list(padded_text_file.get())
        == [
            "a,b,c,d",
            "1,2,3,4",
            "5,6,7,8",
            "9,10,11,12",
            "13,14,15,16",
            "17,18,19,20",
        ]
    )

    assert (
        padded_text_file[:3]
        == list(padded_text_file.get(stop=3))
        == [
            "a,b,c,d",
            "1,2,3,4",
            "5,6,7,8",
        ]
    )

    assert (
        padded_text_file[4:]
        == list(padded_text_file.get(start=4))
        == [
            "13,14,15,16",
            "17,18,19,20",
        ]
    )

    assert (
        padded_text_file[1:4]
        == list(padded_text_file.get(start=1, stop=4))
        == [
            "1,2,3,4",
            "5,6,7,8",
            "9,10,11,12",
        ]
    )

    assert (
        padded_text_file[1:-1]
        == list(padded_text_file.get(start=1, stop=-1))
        == [
            "1,2,3,4",
            "5,6,7,8",
            "9,10,11,12",
            "13,14,15,16",
        ]
    )


def test_padded_text_file(padded_file_path):
    with padded_text_file(padded_file_path) as pdt:
        assert pdt[2] == "5,6,7,8"


def test_len_offset(padded_file_descriptor: IO, padded_file_size: int) -> None:
    padded_text_file = PaddedTextFile(padded_file_descriptor, padded_file_size, 1)
    assert len(padded_text_file) == 5


def test_line_offset(padded_file_descriptor: IO, padded_file_size: int) -> None:
    padded_text_file = PaddedTextFile(padded_file_descriptor, padded_file_size, 1)

    assert padded_text_file[2] == "9,10,11,12"
    assert padded_text_file[-1] == "17,18,19,20"

    with pytest.raises(IndexError):
        padded_text_file[5]

    with pytest.raises(IndexError):
        padded_text_file[-6]


def test_slice_offset(padded_file_descriptor: IO, padded_file_size: int) -> None:
    padded_text_file = PaddedTextFile(padded_file_descriptor, padded_file_size, 1)

    assert (
        padded_text_file[:]
        == padded_text_file[-1000:1000]
        == list(padded_text_file.get())
        == [
            "1,2,3,4",
            "5,6,7,8",
            "9,10,11,12",
            "13,14,15,16",
            "17,18,19,20",
        ]
    )

    assert (
        padded_text_file[:3]
        == padded_text_file[-5:3]
        == list(padded_text_file.get(stop=3))
        == ["1,2,3,4", "5,6,7,8", "9,10,11,12"]
    )

    assert (
        padded_text_file[4:]
        == padded_text_file[-1:]
        == list(padded_text_file.get(start=4))
        == [
            "17,18,19,20",
        ]
    )

    assert (
        padded_text_file[1:4]
        == list(padded_text_file.get(start=1, stop=4))
        == ["5,6,7,8", "9,10,11,12", "13,14,15,16"]
    )

    assert (
        padded_text_file[1:-1]
        == list(padded_text_file.get(start=1, stop=-1))
        == [
            "5,6,7,8",
            "9,10,11,12",
            "13,14,15,16",
        ]
    )


def test_offset_max(padded_file_descriptor: IO, padded_file_size: int) -> None:
    padded_text_file = PaddedTextFile(padded_file_descriptor, padded_file_size, 6)
    assert len(padded_text_file) == 0

    with pytest.raises(IndexError):
        padded_text_file[0]


def test_offset_out_of_bound(padded_file_descriptor: IO, padded_file_size: int) -> None:
    with pytest.raises(OffsetError):
        PaddedTextFile(padded_file_descriptor, padded_file_size, -1)

    with pytest.raises(OffsetError):
        PaddedTextFile(padded_file_descriptor, padded_file_size, 7)


def test_splitted_padded_text_file_offset_0(
    splitted_padded_csv_files_descriptor_size: List[Tuple[IO, int]]
):

    splitted_padded_text_file = SplittedPaddedTextFile(
        splitted_padded_csv_files_descriptor_size, offset=0
    )

    assert len(splitted_padded_text_file) == 6

    assert splitted_padded_text_file[:] == [
        "a,b,c,d",
        "1,2,3,4",
        "5,6,7,8",
        "9,10,11,12",
        "13,14,15,16",
        "17,18,19,20",
    ]

    assert splitted_padded_text_file[-1] == "17,18,19,20"
    assert splitted_padded_text_file[-1:] == ["17,18,19,20"]


def test_splitted_padded_text_file_offset_1(
    splitted_padded_csv_files_descriptor_size: List[Tuple[IO, int]]
):

    splitted_padded_text_file = SplittedPaddedTextFile(
        splitted_padded_csv_files_descriptor_size, offset=1
    )

    assert len(splitted_padded_text_file) == 5

    assert splitted_padded_text_file[:] == [
        "1,2,3,4",
        "5,6,7,8",
        "9,10,11,12",
        "13,14,15,16",
        "17,18,19,20",
    ]

    assert splitted_padded_text_file[-1] == "17,18,19,20"
    assert splitted_padded_text_file[-1:] == ["17,18,19,20"]



from pathlib import Path
from typing import IO

import pytest
from pytest import fixture

from ..padded_csv_file import ColumnNotFoundError, _PaddedCSVFile, padded_csv_file
from . import assets


@fixture
def padded_file_path() -> Path:
    return Path(assets.__file__).parent / "padded.csv"


@fixture
def padded_file_descriptor(padded_file_path: Path) -> IO:
    return padded_file_path.open()


@fixture
def padded_file_size(padded_file_path: Path) -> int:
    return padded_file_path.stat().st_size


def test_bad_columns(padded_file_descriptor: IO, padded_file_size: int) -> None:
    with pytest.raises(ColumnNotFoundError):
        _PaddedCSVFile(
            [(padded_file_descriptor, padded_file_size)], [("not_present", int)]
        )


def test_empy_column_and_type(
    padded_file_descriptor: IO, padded_file_size: int
) -> None:
    with pytest.raises(ValueError):
        _PaddedCSVFile([(padded_file_descriptor, padded_file_size)], [])


def test_len(padded_file_descriptor: IO, padded_file_size: int) -> None:
    padded_csv_file = _PaddedCSVFile(
        [(padded_file_descriptor, padded_file_size)], [("d", int), ("b", int)]
    )

    assert len(padded_csv_file) == 5


def test_line(padded_file_descriptor: IO, padded_file_size: int) -> None:
    padded_csv_file = _PaddedCSVFile(
        [(padded_file_descriptor, padded_file_size)], [("d", int), ("b", int)]
    )

    assert padded_csv_file[0] == [4, 2]
    assert padded_csv_file[-1] == [20, 18]


def test_line_one(padded_file_descriptor: IO, padded_file_size: int) -> None:
    padded_csv_file = _PaddedCSVFile(
        [(padded_file_descriptor, padded_file_size)], [("d", int)]
    )

    assert padded_csv_file[0] == [4]
    assert padded_csv_file[-1] == [20]


def test_line_one_unwrapped(padded_file_descriptor: IO, padded_file_size: int) -> None:
    padded_csv_file = _PaddedCSVFile(
        [(padded_file_descriptor, padded_file_size)],
        [("d", int)],
        unwrap_if_one_column=True,
    )

    assert padded_csv_file[0] == 4
    assert padded_csv_file[-1] == 20


def test_slice(padded_file_descriptor: IO, padded_file_size: int) -> None:
    padded_csv_file = _PaddedCSVFile(
        [(padded_file_descriptor, padded_file_size)], [("d", int), ("b", int)]
    )

    assert (
        padded_csv_file[:]
        == list(padded_csv_file.get())
        == [[4, 2], [8, 6], [12, 10], [16, 14], [20, 18]]
    )

    assert (
        padded_csv_file[:3]
        == list(padded_csv_file.get(stop=3))
        == [[4, 2], [8, 6], [12, 10]]
    )

    assert (
        padded_csv_file[3:]
        == list(padded_csv_file.get(start=3))
        == [[16, 14], [20, 18]]
    )

    assert (
        padded_csv_file[1:4]
        == list(padded_csv_file.get(start=1, stop=4))
        == [[8, 6], [12, 10], [16, 14]]
    )

    assert (
        padded_csv_file[1:-1]
        == list(padded_csv_file.get(start=1, stop=-1))
        == [[8, 6], [12, 10], [16, 14]]
    )


def test_padded_csv_file(padded_file_path):
    with padded_csv_file(padded_file_path, [("d", int), ("b", int)]) as pcf:
        assert pcf[2] == [12, 10]


def test_slice_one(padded_file_descriptor: IO, padded_file_size: int) -> None:
    padded_csv_file = _PaddedCSVFile(
        [(padded_file_descriptor, padded_file_size)], [("d", int)]
    )

    assert (
        padded_csv_file[:]
        == list(padded_csv_file.get())
        == [[4], [8], [12], [16], [20]]
    )


def test_slice_one_unwrapped(padded_file_descriptor: IO, padded_file_size: int) -> None:
    padded_csv_file = _PaddedCSVFile(
        [(padded_file_descriptor, padded_file_size)],
        [("d", int)],
        unwrap_if_one_column=True,
    )

    assert padded_csv_file[:] == list(padded_csv_file.get()) == [4, 8, 12, 16, 20]



import filecmp
import os
from pathlib import Path

from pytest import fixture

from ..pad_and_sample import (
    are_files_fully_sampled,
    compute_chunks,
    pad,
    pad_and_sample,
    pseudo_hash,
    sample,
    sample_sampled,
)
from ..tests import assets


@fixture
def padded_file_path() -> Path:
    return Path(assets.__file__).parent / "padded.csv"


@fixture
def padded_0_1_file_path() -> Path:
    return Path(assets.__file__).parent / "padded_0_1.csv"


@fixture
def padded_2_4_file_path() -> Path:
    return Path(assets.__file__).parent / "padded_2_4.csv"


@fixture
def padded_5_file_path() -> Path:
    return Path(assets.__file__).parent / "padded_5.csv"


@fixture
def sampled_file_path() -> Path:
    return Path(assets.__file__).parent / "sampled.csv"


@fixture
def sampled_0_1_file_path() -> Path:
    return Path(assets.__file__).parent / "sampled_0_1.csv"


@fixture
def sampled_2_4_file_path() -> Path:
    return Path(assets.__file__).parent / "sampled_2_4.csv"


@fixture
def sampled_5_file_path() -> Path:
    return Path(assets.__file__).parent / "sampled_5.csv"


@fixture
def not_padded_file_path() -> Path:
    path = Path(assets.__file__).parent / "not_padded.csv"
    os.utime(path, (42, 42))
    return path


@fixture
def hashed() -> Path:
    return Path(assets.__file__).parent / "8bed00c4529bfd12bd70678a71eaf5af"


@fixture
def _0(hashed: Path) -> Path:
    return hashed / "0"


@fixture
def _1(hashed: Path) -> Path:
    return hashed / "1"


@fixture
def _2(hashed: Path) -> Path:
    return hashed / "2"


@fixture
def splitted_padded_csv_path() -> Path:
    return Path(assets.__file__).parent / "splitted_padded_csv"


@fixture
def fully_splitted_padded_csv_path() -> Path:
    return Path(assets.__file__).parent / "fully_splitted_padded_csv"


def test_pad(tmpdir, not_padded_file_path, padded_file_path):
    pad(not_padded_file_path, Path(tmpdir) / "padded.csv")
    assert filecmp.cmp(tmpdir / "padded.csv", padded_file_path)


def test_pad_0_1(tmpdir, not_padded_file_path, padded_0_1_file_path):
    pad(not_padded_file_path, Path(tmpdir) / "padded.csv", 0, 16)
    assert filecmp.cmp(tmpdir / "padded.csv", padded_0_1_file_path)


def test_pad_2_4(tmpdir, not_padded_file_path, padded_2_4_file_path):
    pad(not_padded_file_path, Path(tmpdir) / "padded.csv", 16, 47)
    assert filecmp.cmp(tmpdir / "padded.csv", padded_2_4_file_path)


def test_pad_5(tmpdir, not_padded_file_path, padded_5_file_path):
    pad(not_padded_file_path, Path(tmpdir) / "padded.csv", 47, 59)
    assert filecmp.cmp(tmpdir / "padded.csv", padded_5_file_path)


def test_pseudo_hash(tmp_path: Path):
    path = tmp_path / "file.txt"
    with path.open("w") as file_descriptor:
        file_descriptor.write("qwerty")

    os.utime(path, (42, 42))

    assert pseudo_hash(path) == "46141dc64506ddc74fcf3cae348e0d77"
    assert pseudo_hash(path, "12345") == "ece35c8647c0fe39e450443c9514b412"


def test_compute_chunks(not_padded_file_path: Path):
    assert compute_chunks(not_padded_file_path, 1) == [
        (0, 59),
    ]

    assert compute_chunks(not_padded_file_path, 2) == [
        (0, 35),
        (35, 59),
    ]

    assert compute_chunks(not_padded_file_path, 4) == [
        (0, 16),
        (16, 35),
        (35, 47),
        (47, 59),
    ]

    assert compute_chunks(not_padded_file_path, 10) == [
        (0, 8),
        (8, 16),
        (16, 24),
        (24, 35),
        (35, 47),
        (47, 59),
    ]


def test_compute_chunks_padded(padded_file_path: Path):
    assert compute_chunks(padded_file_path, 1) == [
        (0, 72),
    ]

    assert compute_chunks(padded_file_path, 2) == [
        (0, 36),
        (36, 72),
    ]

    assert compute_chunks(padded_file_path, 3) == [
        (0, 24),
        (24, 48),
        (48, 72),
    ]

    assert compute_chunks(padded_file_path, 6) == [
        (0, 12),
        (12, 24),
        (24, 36),
        (36, 48),
        (48, 60),
        (60, 72),
    ]


def test_are_files_fully_sampled(
    splitted_padded_csv_path: Path, fully_splitted_padded_csv_path: Path
):
    assert not are_files_fully_sampled(splitted_padded_csv_path)
    assert are_files_fully_sampled(fully_splitted_padded_csv_path)


def test_sample(tmpdir, not_padded_file_path, sampled_file_path):
    sample(
        not_padded_file_path,
        tmpdir / "output.csv",
        "a",
        2,
    )

    assert filecmp.cmp(tmpdir / "output.csv", sampled_file_path)


def test_sample_0_1(tmpdir, not_padded_file_path, sampled_0_1_file_path):
    sample(not_padded_file_path, tmpdir / "output.csv", "a", 2, 0, 16)

    assert filecmp.cmp(tmpdir / "output.csv", sampled_0_1_file_path)


def test_sample_2_4(tmpdir, not_padded_file_path, sampled_2_4_file_path):
    sample(not_padded_file_path, tmpdir / "output.csv", "a", 2, 16, 47)

    assert filecmp.cmp(tmpdir / "output.csv", sampled_2_4_file_path)


def test_sample_5(tmpdir, not_padded_file_path, sampled_5_file_path):
    sample(not_padded_file_path, tmpdir / "output.csv", "a", 2, 47, 59)

    assert filecmp.cmp(tmpdir / "output.csv", sampled_5_file_path)


def test_sample_sampled_full_header(tmpdir, sampled_0_1_file_path):
    sample_sampled(sampled_0_1_file_path, Path(tmpdir) / "output.csv", 2, True)

    assert filecmp.cmp(tmpdir / "output.csv", sampled_0_1_file_path)


def test_sample_sampled_full_no_header(tmpdir, sampled_5_file_path):
    sample_sampled(sampled_5_file_path, Path(tmpdir) / "output.csv", 2, False)

    assert filecmp.cmp(tmpdir / "output.csv", sampled_5_file_path)


def test_pad_and_sample(
    tmp_path: Path, not_padded_file_path: Path, _0: Path, _1: Path, _2: Path
):
    assert pad_and_sample(not_padded_file_path, tmp_path, "a", 2)

    assert filecmp.cmp(
        tmp_path / "8bed00c4529bfd12bd70678a71eaf5af" / "0" / "0.csv", _0 / "0.csv"
    )

    assert filecmp.cmp(
        tmp_path / "8bed00c4529bfd12bd70678a71eaf5af" / "0" / "1.csv", _0 / "1.csv"
    )

    assert filecmp.cmp(
        tmp_path / "8bed00c4529bfd12bd70678a71eaf5af" / "1" / "0.csv", _1 / "0.csv"
    )

    assert filecmp.cmp(
        tmp_path / "8bed00c4529bfd12bd70678a71eaf5af" / "1" / "1.csv", _1 / "1.csv"
    )

    assert filecmp.cmp(
        tmp_path / "8bed00c4529bfd12bd70678a71eaf5af" / "2" / "0.csv", _2 / "0.csv"
    )

    assert filecmp.cmp(
        tmp_path / "8bed00c4529bfd12bd70678a71eaf5af" / "2" / "1.csv", _2 / "1.csv"
    )

    assert not pad_and_sample(not_padded_file_path, tmp_path, "a", 2)



from enum import Enum
from typing import Any, Dict, List, Optional, Set

from pydantic import BaseModel, Field, constr, validator


class Color(str, Enum):
    Aqua = "aqua"
    Black = "black"
    Blue = "blue"
    Fuchsia = "fuchsia"
    Gray = "gray"
    Green = "green"
    Lime = "lime"
    Maroon = "maroon"
    Navy = "navy"
    Olive = "olive"
    Orange = "orange"
    Purple = "purple"
    Red = "red"
    Silver = "silver"
    Teal = "teal"
    White = "white"
    Yellow = "yellow"


COLOR_NAME_TO_HEXA = {
    # https://clrs.cc/
    Color.Aqua: "#7FDBFF",
    Color.Black: "#111111",
    Color.Blue: "#0074D9",
    Color.Fuchsia: "F012BE",
    Color.Gray: "#AAAAAA",
    Color.Green: "#2ECC40",
    Color.Lime: "#01FF70",
    Color.Maroon: "85144b",
    Color.Navy: "#001f3f",
    Color.Olive: "#3D9970",
    Color.Orange: "#FF851B",
    Color.Purple: "#B10DC9",
    Color.Red: "#FF4136",
    Color.Silver: "#DDDDDD",
    Color.Teal: "#39CCCC",
    Color.White: "#FFFFFF",
    Color.Yellow: "#FFDC00",
}


class Configuration(BaseModel):
    class General(BaseModel):
        variable: str
        label: Optional[str]
        unit: Optional[str]
        as_datetime: bool = Field(False, alias="asDateTime")
        date_time_formats: Optional[List[str]] = Field(None, alias="dateTimeFormats")

    class LayoutItem(BaseModel):
        position: constr(regex=r"^[0-9]+-[0-9]+$")  # type: ignore
        title: Optional[str]
        label: Optional[str]
        unit: Optional[str]

        # Computed
        x: int = 0
        y: int = 0

        @validator("x", always=True)
        def set_x(cls, _, values: Dict[str, Any]) -> int:
            x, _ = values["position"].split("-")
            return int(x)

        @validator("y", always=True)
        def set_y(cls, _, values: Dict[str, Any]) -> int:
            _, y = values["position"].split("-")
            return int(y)

    class Curve(BaseModel):
        variable: str
        position: constr(regex=r"^[0-9]+-[0-9]+$") = "1-1"  # type: ignore
        color: Color = Color.Yellow

        # Computed
        x: int = 0
        y: int = 0

        @validator("x", always=True)
        def set_x(cls, _, values: Dict[str, Any]) -> int:
            x, _ = values["position"].split("-")
            return int(x)

        @validator("y", always=True)
        def set_y(cls, _, values: Dict[str, Any]) -> int:
            _, y = values["position"].split("-")
            return int(y)

    general: General
    layout: Optional[List[LayoutItem]]
    curves: List[Curve]
    variables: Set[str] = set()

    @validator("variables", always=True)
    def set_variables(cls, value, values: Dict[str, Any]) -> Set[str]:
        if "curves" not in values:
            return value

        return {curve.variable for curve in values["curves"]}



from contextlib import ExitStack, contextmanager
from datetime import datetime
from glob import glob
from pathlib import Path
from typing import Any, Dict, Iterator, List, Set, Tuple, Union

from pydantic import BaseModel

from .sorted_padded_csv_file import _SortedPaddedCSVFile, sorted_padded_csv_file


class Selected(BaseModel):
    class Y(BaseModel):
        mins: List[float]
        maxs: List[float]

    xs: Union[List[float], List[datetime]]
    name_to_y: Dict[str, Y]


class _Selector:
    """This class is a helper to get the requested data as closest as possible to a
    given resolution.

    Usage:
    ======

    File 0.csv:
    -----------
    a,b,c,d
    1,2,3,4
    5,6,7,8
    9,10,11,12
    13,14,15,16
    17,18,19,20

    File 1.csv:
    -----------
    a,b_min,b_max,d_min,d_max
    1,2.0,6.0,4.0,8.0
    9,10.0,14.0,12.0,16.0
    17,18.0,18.0,20.0,20.0

    File 2.csv:
    -----------
    a,b_min,b_max,d_min,d_max
    1,2.0,14.0,4.0,16.0
    17,18.0,18.0,20.0,20.0

    File 3.csv:
    -----------
    a,b_min,b_max,d_min,d_max
    1,2.0,18.0,4.0,20.0

    sel = _Selector(
            <spcf corresponding to `0.csv`>,
            [("b", float), ("d", float)],
            {
              <spcf corresponding to `2.csv`>,
              <spcf corresponding to `3.csv`>,
              <spcf corresponding to `4.csv`>
            },
            [("b_min", float), ("b_max", float), ("d_min", float), ("d_max", float)]
          )

    sel[::6] == Selected(
                xs=[1, 5, 9, 13, 17],
                name_to_y={
                  "b": Selected.Y(mins=[2, 6, 10, 14, 18], maxs=[2, 6, 10, 14, 18]),
                  "d": Selected.Y(mins=[4, 8, 12, 16, 20], maxs=[4, 8, 12, 16, 20]),
                },
              )

    sel[5:13:6] == sel[4.5:13.5:6]
              == Selected(
                   xs=[5, 9, 13],
                   name_to_y={
                     "b": Selected.Y(mins=[6, 10, 14], maxs=[6, 10, 14]),
                     "d": Selected.Y(mins=[8, 12, 16], maxs=[8, 12, 16]),
                   },
                 )

    sel[::4] == Selected(
                xs=[1, 5, 9, 13, 17],
                name_to_y={
                  "b": Selected.Y(mins=[2, 6, 10, 14, 18], maxs=[2, 6, 10, 14, 18]),
                  "d": Selected.Y(mins=[4, 8, 12, 16, 20], maxs=[4, 8, 12, 16, 20]),
                },
              )

    sel[5:13:4] == sel[4.5:13.5:4]
              == Selected(
                   xs=[5, 9, 13],
                   name_to_y={
                     "b": Selected.Y(mins=[6, 10, 14], maxs=[6, 10, 14]),
                     "d": Selected.Y(mins=[8, 12, 16], maxs=[8, 12, 16]),
                   },
                 )

    sel[::3] == Selected(
                xs=[1, 9, 17],
                name_to_y={
                  "b": Selected.Y(mins=[2, 10, 18], maxs=[6, 14, 18]),
                  "d": Selected.Y(mins=[4, 12, 20], maxs=[8, 16, 20]),
                },
              )

    sel[1:9:2] == sel[0.5:9.5:2]
             == Selected(
                  xs=[1, 9],
                  name_to_y={
                    "b": Selected.Y(mins=[2, 10], maxs=[6, 14]),
                    "d": Selected.Y(mins=[4, 12], maxs=[8, 16]),
                  },
                )
    """

    def __init__(
        self,
        spcf: _SortedPaddedCSVFile,
        ys: List[str],
        sampled_spcfs: Set[_SortedPaddedCSVFile],
        sampled_ys: List[str],
    ) -> None:
        """Initializer:

        spcf                : A (non sampled) Sorted Padded CSV file
        ys_and_type         : Y names and Y types of `spcf` file

        sampled_spcfs       : A set of sampled Sorted Padded CSV files corresponding to
                              spcf

        sampled_ys_and_types: Y names an Y types corresponding to sampled_spcfs files
                              Note: This value has to be the same for all sampled_spcfs
        """
        self.__spcf = spcf
        self.__sampled_spcfs = sampled_spcfs
        self.__all_spcfs = self.__sampled_spcfs.union({self.__spcf})

        self.__y_names = ys
        self.__sampled_y_names = sampled_ys

    def __get_nb_lines_between(
        self, start: Any, stop: Any
    ) -> Dict[_SortedPaddedCSVFile, int]:
        """Get the number of lines between `start` and `stop` for each Sorted Padded CSV
        file."""
        return {
            spcf: spcf.number_of_lines_between(start, stop) for spcf in self.__all_spcfs
        }

    def __get_max_resolution_lines_between(
        self, start: Any, stop: Any, resolution: int
    ) -> _SortedPaddedCSVFile:
        """Return the smallest Sorted Padded CSV file where the number of lines between
        `start` and `stop` is higher than `resolution`"""
        spcf_to_nb_lines = self.__get_nb_lines_between(start, stop)
        spcf_max = max(spcf_to_nb_lines, key=spcf_to_nb_lines.get)  # type: ignore

        spcf_to_nb_lines_filtered = {
            path: nb_lines
            for path, nb_lines in spcf_to_nb_lines.items()
            if nb_lines >= resolution
        }

        return (
            min(spcf_to_nb_lines_filtered, key=spcf_to_nb_lines_filtered.get)  # type: ignore
            if len(spcf_to_nb_lines_filtered) > 0
            else spcf_max
        )

    def __getitem__(self, x_or_slice: Union[Any, slice]) -> Selected:
        """Return a Selected object where the number of lines are as close as (but
        always greater than) the resolution."""
        assert isinstance(x_or_slice, slice), "Only slice is supported for `x_or_slice`"

        start, stop, step = x_or_slice.start, x_or_slice.stop, x_or_slice.step

        if step is None:
            raise ValueError("Step of slice has to be defined")

        spcf = self.__get_max_resolution_lines_between(start, stop, step)
        xs, y_in_rows = zip(*spcf[start:stop])
        y_in_columns = list(zip(*y_in_rows))

        if spcf == self.__spcf:
            name_to_y = {
                y_name: Selected.Y(mins=y_in_column, maxs=y_in_column)
                for y_name, y_in_column in zip(self.__y_names, y_in_columns)
            }
        else:
            sampled_name_to_y = {
                name: y for name, y in zip(self.__sampled_y_names, y_in_columns)
            }

            name_to_y = {
                y_name: Selected.Y(
                    mins=sampled_name_to_y[f"{y_name}_min"],
                    maxs=sampled_name_to_y[f"{y_name}_max"],
                )
                for y_name in self.__y_names
            }

        return Selected(xs=xs, name_to_y=name_to_y)


@contextmanager
def selector(
    dir_path: Path,
    x_and_type: Tuple[str, type],
    ys: List[str],
) -> Iterator[_Selector]:
    """Select the sampled file matching as close as possible a given resolution.

    dir_path   : Directory where all the files (sampled and non sampled) are located.
                 Non sampled path name's HAS to be `0.csv`

    x_and_type : Name and the type of X value
    ys: ys name

    Usage:
    ======

    File 0.csv:
    -----------
    a,b,c,d
    1,2,3,4
    5,6,7,8
    9,10,11,12
    13,14,15,16
    17,18,19,20

    File 1.csv:
    -----------
    a,b_min,b_max,d_min,d_max
    1,2.0,6.0,4.0,8.0
    9,10.0,14.0,12.0,16.0
    17,18.0,18.0,20.0,20.0

    File 2.csv:
    -----------
    a,b_min,b_max,d_min,d_max
    1,2.0,14.0,4.0,16.0
    17,18.0,18.0,20.0,20.0

    File 3.csv:
    -----------
    a,b_min,b_max,d_min,d_max
    1,2.0,18.0,4.0,20.0

    dir_path
    |
    |- 0.csv
    |- 1.csv
    |- 2.csv
    |- 3.csv

    with selector(dir_path, ("a", int), ["b", "d"]) as sel:
        sel[::100] == Selected(
                    xs=[1, 5, 9, 13, 17],
                    name_to_y={
                    "b": Selected.Y(mins=[2, 6, 10, 14, 18], maxs=[2, 6, 10, 14, 18]),
                    "d": Selected.Y(mins=[4, 8, 12, 16, 20], maxs=[4, 8, 12, 16, 20]),
                    },
                )

        sel[5:13:100] == sel[4.5:13.5:100]
                  == Selected(
                       xs=[5, 9, 13],
                       name_to_y={
                         "b": Selected.Y(mins=[6, 10, 14], maxs=[6, 10, 14]),
                         "d": Selected.Y(mins=[8, 12, 16], maxs=[8, 12, 16]),
                       },
                     )

        sel[::4] == Selected(
                    xs=[1, 5, 9, 13, 17],
                    name_to_y={
                      "b": Selected.Y(mins=[2, 6, 10, 14, 18], maxs=[2, 6, 10, 14, 18]),
                      "d": Selected.Y(mins=[4, 8, 12, 16, 20], maxs=[4, 8, 12, 16, 20]),
                    },
                  )

        sel[5:13:4] == sel[4.5:13.5:4]
                  == Selected(
                       xs=[5, 9, 13],
                       name_to_y={
                         "b": Selected.Y(mins=[6, 10, 14], maxs=[6, 10, 14]),
                         "d": Selected.Y(mins=[8, 12, 16], maxs=[8, 12, 16]),
                       },
                     )

        sel[::3] == Selected(
                    xs=[1, 9, 17],
                    name_to_y={
                      "b": Selected.Y(mins=[2, 10, 18], maxs=[6, 14, 18]),
                      "d": Selected.Y(mins=[4, 12, 20], maxs=[8, 16, 20]),
                    },
                  )

        sel[1:9:2] == sel[0.5:9.5]
                 == Selected(
                      xs=[1, 9],
                      name_to_y={
                        "b": Selected.Y(mins=[2, 10], maxs=[6, 14]),
                        "d": Selected.Y(mins=[4, 12], maxs=[8, 16]),
                      },
                    )

    """
    sampled_paths = [
        path for path in dir_path.iterdir() if path.name not in ("0", "SUCCESS")
    ]

    sampled_ys = [
        item for sublist in [[f"{y}_min", f"{y}_max"] for y in ys] for item in sublist
    ]

    with ExitStack() as stack:
        spcf = stack.enter_context(
            sorted_padded_csv_file(dir_path / "0", x_and_type, ys)
        )

        sampled_spcfs = {
            stack.enter_context(
                sorted_padded_csv_file(sampled_path, x_and_type, sampled_ys)
            )
            for sampled_path in sampled_paths
        }

        yield _Selector(spcf, ys, sampled_spcfs, sampled_ys)



from contextlib import contextmanager
from pathlib import Path
from typing import IO, Any, Iterator, List, Optional, Tuple, Union, cast

from .gettable import Gettable
from .padded_text_file import SplittedPaddedTextFile


class ColumnNotFoundError(Exception):
    pass


class _PaddedCSVFile(Gettable):
    """Represent a padded CSV file, where lines are reachable with O(1) complexity.

    A padded CSV file is a CSV file where all lines have exactly the same length.
    In general, lines are right padded with white spaces.
    The last line MUST also contain a carriage return.

    Only line(s) you request will be load in memory.

    Usage:
    padded_csv_file = _PaddedCSVFile(<files_descriptor_and_size>,
                                     <file_zise>,
                                     <column_and_type_tuples>
                                    )

    Example: With the following file represented by <files_descriptor_and_size>:
    a,b,c,d
    1,2,3,4
    5,6,7,8
    9,10,11,12
    13,14,15,16
    17,18,19,20

    padded_csv_file = _PaddedCSVFile(<files_descriptor_and_size>,
                                     <file_size>,
                                     [("d", int), ("b", int)]
                                    )

    # Get the number of lines
    len(padded_csv_file) # = 5

    # Get the third line of the file
    padded_csv_file[2] # = [12, 10]

    # Get the last line of the file
    padded_csv_file[-1] # = [20, 18]

    # Get an iterator on lines between the third line (included) and the last line
    # (excluded)
    padded_csv_file.get(start=2, stop=-1)

    # Get all lines between the third line (included) and the last line (excluded)
    # Warning: All lines in the selected range will be loaded into memory.
    #          For example: padded_csv_file[:] will load all the file in memory.
    #          If possible, use padded_csv_file.get(start=a, stop=b) instead of
    #                           padded_csv_file[a, b]
    padded_csv_file[2:-1] # = [[12, 10], [16, 14]]
    """

    def __init__(
        self,
        files_descriptor_and_size: List[Tuple[IO, int]],
        columns_and_types: List[Tuple[str, type]],
        unwrap_if_one_column=False,
    ) -> None:
        """Constructor.

        files_descriptor_and_size:
            A liste of tuples like:
                - The file descriptor pointing to the padded CSV file
                - The size (in bytes) of th padded CSV file pointed by the file
                  descriptor
        columns_and_types: A list of tuples where each tuple has:
                           - The name of the column
                           - The type of the column
        unwrap_if_one_column: Unwrap if only one column unwrap result.
                              Exemple: Instead of returning [[4], [5], [2]] return
                                       [4, 5, 2]

        If at least one line of the file pointed by `file_descriptor` has not the same
        length than others, a `TextFileNotPaddedError` is raised.
        """
        padded_text_file = SplittedPaddedTextFile(files_descriptor_and_size, offset=0)
        header_line = cast(str, padded_text_file[0])
        headers = header_line.split(",")

        if columns_and_types == []:
            raise ValueError("`column_and_type` is an empty list")

        columns, _ = zip(*columns_and_types)

        if not set(columns) <= set(headers):
            raise ColumnNotFoundError(
                "At least one column specified in `column_to_type` in not present in "
                "the file"
            )

        header_to_index = {header: index for index, header in enumerate(headers)}

        self.__column_indexes_type = [
            (header_to_index[column], type) for column, type in columns_and_types
        ]

        self.__padded_text_file = SplittedPaddedTextFile(
            files_descriptor_and_size, offset=1
        )
        _, *others = columns_and_types
        self.__has_to_unwrap = unwrap_if_one_column and others == []

    def __len__(self):
        """Return the number of lines of the file (excluding the header)."""
        return len(self.__padded_text_file)

    def __getitem__(
        self, line_number_or_slice: Union[int, slice]
    ) -> Union[Any, List, List[List]]:
        """Get given values or a given slice of values.

        line_number_or_slice: The line number or the slice where values will be
                              retrieved
        """

        def handle_line_number(line_number: int) -> Union[Any, List]:
            line = cast(str, self.__padded_text_file[line_number])
            items = line.split(",")

            return self.__unwrap_if_needed_single(
                [type(items[index]) for index, type in self.__column_indexes_type]
            )

        def handle_slice(slice: slice) -> List[Union[Any, List]]:
            return self.__unwrap_if_needed_multi(
                [
                    [
                        type(items.split(",")[index])
                        for index, type in self.__column_indexes_type
                    ]
                    for items in self.__padded_text_file[slice]
                ]
            )

        if isinstance(line_number_or_slice, int):
            return handle_line_number(line_number_or_slice)
        elif isinstance(line_number_or_slice, slice):
            return handle_slice(line_number_or_slice)

    def __unwrap_if_needed_single(self, items: List) -> Union[List, Any]:
        if self.__has_to_unwrap:
            item, *trash = items
            assert trash == []
            return item
        return items

    def __unwrap_if_needed_multi(self, items: List[List]) -> List:
        return (
            [item for sublist in items for item in sublist]
            if self.__has_to_unwrap
            else items
        )

    def get(
        self, start: Optional[int] = None, stop: Optional[int] = None
    ) -> Iterator[List]:
        """Return an iterator on a given slice of lines.

        start: The first line of slice (included)
        stop : The last line of slice (excluded)
        """
        for line in self.__padded_text_file.get(start, stop):
            items = line.split(",")
            toto = [type(items[index]) for index, type in self.__column_indexes_type]
            yield self.__unwrap_if_needed_single(toto)


@contextmanager
def padded_csv_file(
    path: Path, columns_and_types: List[Tuple[str, type]]
) -> Iterator[_PaddedCSVFile]:
    """Represent a padded CSV file, where lines are reachable with O(1) complexity.

    A padded CSV file is a CSV file where all lines have exactly the same length.
    In general, lines are right padded with white spaces.
    The last line MUST also contain a carriage return.

    Only line(s) you request will be load in memory.

    Usage:
    with padded_csv_file(<file_path>, <columns_and_types>) as pcf:
        ...

    Example: With the following file represented by <file_descriptor>:
    a,b,c,d
    1,2,3,4
    5,6,7,8
    9,10,11,12
    13,14,15,16
    17,18,19,20

    with padded_csv_file(<file_path>, [("d", int), ("b", int)]) as pcf:
        # Get the number of lines
        len(pcf) # = 5

        # Get the third line of the file
        pcf[2] # = [12, 10]

        # Get the last line of the file
        pcf[-1] # = [20, 18]

        # Get an iterator on lines between the third line (included) and the last line
        # (excluded)
        pcf.get(start=2, stop=-1)

        # Get all lines between the third line (included) and the last line (excluded)
        # Warning: All lines in the selected range will be loaded into memory.
        #          For example: padded_csv_file[:] will load all the file in memory.
        #          If possible, use pcf.get(start=a, stop=b) instead of
        #                           pcf[a, b]
        pcf[2:-1] # = [[12, 10], [16, 14]]
    """
    with path.open() as file_descriptor:
        yield _PaddedCSVFile(
            [(file_descriptor, path.stat().st_size)], columns_and_types
        )



from abc import ABC, abstractmethod
from typing import Any, Iterator, List, Optional, Union


class Gettable(ABC):
    @abstractmethod
    def __len__(self) -> int:
        ...  # pragma: no cover

    @abstractmethod
    def get(self, start: Optional[Any], stop: Optional[Any]) -> Iterator:
        ...  # pragma: no cover

    @abstractmethod
    def __getitem__(self, index_or_slice: Union[Any, slice]) -> Union[Any, List]:
        ...  # pragma: no cover



from typing import Iterable, List, Optional

import pytest

from ..splitted_gettable import SplittedGettable


class GettableList(list):
    def get(self, start: Optional[int], stop: Optional[int]) -> Iterable:
        for index in range(
            start if start is not None else 0,
            stop if stop is not None else len(self),
        ):
            yield self[index]


@pytest.fixture
def gettable_lists() -> List[GettableList]:
    return [
        GettableList(["a", "b", "c"]),
        GettableList(["d"]),
        GettableList(["e", "f", "g", "h"]),
        GettableList(["i", "j"]),
    ]


def test_splitted_iterable_index_offset_0(gettable_lists):
    splitted_iterable = SplittedGettable(gettable_lists, 0)

    assert (len(splitted_iterable)) == 10

    with pytest.raises(IndexError):
        splitted_iterable[-11]

    assert splitted_iterable[0] == splitted_iterable[-10] == "a"
    assert splitted_iterable[1] == splitted_iterable[-9] == "b"
    assert splitted_iterable[2] == splitted_iterable[-8] == "c"
    assert splitted_iterable[3] == splitted_iterable[-7] == "d"
    assert splitted_iterable[4] == splitted_iterable[-6] == "e"
    assert splitted_iterable[5] == splitted_iterable[-5] == "f"
    assert splitted_iterable[6] == splitted_iterable[-4] == "g"
    assert splitted_iterable[7] == splitted_iterable[-3] == "h"
    assert splitted_iterable[8] == splitted_iterable[-2] == "i"
    assert splitted_iterable[9] == splitted_iterable[-1] == "j"

    with pytest.raises(IndexError):
        splitted_iterable[10]


def test_splitted_iterable_slice_offset_0(gettable_lists):
    splitted_iterable = SplittedGettable(gettable_lists, 0)

    assert splitted_iterable[0:0] == []

    assert splitted_iterable[:1] == ["a"]
    assert splitted_iterable[:2] == ["a", "b"]
    assert splitted_iterable[:3] == ["a", "b", "c"]
    assert splitted_iterable[:4] == ["a", "b", "c", "d"]
    assert splitted_iterable[:5] == ["a", "b", "c", "d", "e"]
    assert splitted_iterable[:6] == ["a", "b", "c", "d", "e", "f"]
    assert splitted_iterable[:7] == ["a", "b", "c", "d", "e", "f", "g"]
    assert splitted_iterable[:8] == ["a", "b", "c", "d", "e", "f", "g", "h"]
    assert splitted_iterable[:9] == ["a", "b", "c", "d", "e", "f", "g", "h", "i"]

    assert (
        splitted_iterable[:10]
        == splitted_iterable[:11]
        == splitted_iterable[:]
        == ["a", "b", "c", "d", "e", "f", "g", "h", "i", "j"]
    )

    assert splitted_iterable[-3:] == ["h", "i", "j"]


def test_splitted_iterable_index_offset_1(gettable_lists):
    splitted_iterable = SplittedGettable(gettable_lists, 1)

    assert (len(splitted_iterable)) == 9

    with pytest.raises(IndexError):
        splitted_iterable[-10]

    assert splitted_iterable[0] == splitted_iterable[-9] == "b"
    assert splitted_iterable[1] == splitted_iterable[-8] == "c"
    assert splitted_iterable[2] == splitted_iterable[-7] == "d"
    assert splitted_iterable[3] == splitted_iterable[-6] == "e"
    assert splitted_iterable[4] == splitted_iterable[-5] == "f"
    assert splitted_iterable[5] == splitted_iterable[-4] == "g"
    assert splitted_iterable[6] == splitted_iterable[-3] == "h"
    assert splitted_iterable[7] == splitted_iterable[-2] == "i"
    assert splitted_iterable[8] == splitted_iterable[-1] == "j"

    with pytest.raises(IndexError):
        splitted_iterable[9]


def test_splitted_iterable_slice_offset_1(gettable_lists):
    splitted_iterable = SplittedGettable(gettable_lists, 1)

    assert splitted_iterable[0:0] == []

    assert splitted_iterable[:1] == ["b"]
    assert splitted_iterable[:2] == ["b", "c"]
    assert splitted_iterable[:3] == ["b", "c", "d"]
    assert splitted_iterable[:4] == ["b", "c", "d", "e"]
    assert splitted_iterable[:5] == ["b", "c", "d", "e", "f"]
    assert splitted_iterable[:6] == ["b", "c", "d", "e", "f", "g"]
    assert splitted_iterable[:7] == ["b", "c", "d", "e", "f", "g", "h"]
    assert splitted_iterable[:8] == ["b", "c", "d", "e", "f", "g", "h", "i"]

    assert (
        splitted_iterable[:9]
        == splitted_iterable[:10]
        == splitted_iterable[:]
        == ["b", "c", "d", "e", "f", "g", "h", "i", "j"]
    )

    assert splitted_iterable[-3:] == ["h", "i", "j"]



from pathlib import Path
from typing import IO

import pytest
from pytest import fixture

from ..sorted_padded_csv_file import _SortedPaddedCSVFile, sorted_padded_csv_file
from . import assets


@fixture
def padded_file_path() -> Path:
    return Path(assets.__file__).parent / "padded.csv"


@fixture
def padded_file_descriptor_1(padded_file_path: Path) -> IO:
    return padded_file_path.open()


@fixture
def padded_file_descriptor_2(padded_file_path: Path) -> IO:
    return padded_file_path.open()


@fixture
def padded_file_size(padded_file_path: Path) -> int:
    return padded_file_path.stat().st_size


@fixture
def splitted_padded_csv() -> Path:
    return Path(assets.__file__).parent / "splitted_padded_csv"


def test_len(
    padded_file_descriptor_1: IO, padded_file_descriptor_2: IO, padded_file_size: int
) -> None:
    sorted_padded_csv_file = _SortedPaddedCSVFile(
        [(padded_file_descriptor_1, padded_file_descriptor_2, padded_file_size)],
        ("c", int),
        ["d", "b"],
    )

    assert len(sorted_padded_csv_file) == 5


def test_line(
    padded_file_descriptor_1: IO, padded_file_descriptor_2: IO, padded_file_size: int
) -> None:
    sorted_padded_csv_file = _SortedPaddedCSVFile(
        [(padded_file_descriptor_1, padded_file_descriptor_2, padded_file_size)],
        ("c", int),
        ["d", "b"],
    )

    assert sorted_padded_csv_file[7] == (7, [8, 6])

    with pytest.raises(IndexError):
        sorted_padded_csv_file[8]


def test_slice(
    padded_file_descriptor_1: IO, padded_file_descriptor_2: IO, padded_file_size: int
) -> None:
    sorted_padded_csv_file = _SortedPaddedCSVFile(
        [(padded_file_descriptor_1, padded_file_descriptor_2, padded_file_size)],
        ("c", int),
        ["d", "b"],
    )

    assert (
        sorted_padded_csv_file[:]
        == list(sorted_padded_csv_file.get())
        == [
            (3, [4, 2]),
            (7, [8, 6]),
            (11, [12, 10]),
            (15, [16, 14]),
            (19, [20, 18]),
        ]
    )

    assert (
        sorted_padded_csv_file[:3]
        == sorted_padded_csv_file[:3.5]  # type:ignore
        == list(sorted_padded_csv_file.get(stop=3))
        == list(sorted_padded_csv_file.get(stop=3.5))
        == [(3, [4, 2])]
    )

    assert (
        sorted_padded_csv_file[:7]
        == sorted_padded_csv_file[:7.5]  # type: ignore
        == list(sorted_padded_csv_file.get(stop=7))
        == list(sorted_padded_csv_file.get(stop=7.5))
        == [(3, [4, 2]), (7, [8, 6])]
    )

    assert (
        sorted_padded_csv_file[7:15]
        == sorted_padded_csv_file[6.5:15.5]  # type: ignore
        == list(sorted_padded_csv_file.get(start=7, stop=15))
        == list(sorted_padded_csv_file.get(start=6.5, stop=15.5))
        == [(7, [8, 6]), (11, [12, 10]), (15, [16, 14])]
    )

    assert (
        sorted_padded_csv_file[15:19]
        == sorted_padded_csv_file[14.5:42]  # type: ignore
        == list(sorted_padded_csv_file.get(start=15, stop=19))
        == list(sorted_padded_csv_file.get(start=14.5, stop=42))
        == [(15, [16, 14]), (19, [20, 18])]
    )


def test_number_of_lines_between(
    padded_file_descriptor_1: IO, padded_file_descriptor_2: IO, padded_file_size: int
) -> None:
    sorted_padded_csv_file = _SortedPaddedCSVFile(
        [(padded_file_descriptor_1, padded_file_descriptor_2, padded_file_size)],
        ("c", int),
        ["d", "b"],
    )

    assert sorted_padded_csv_file.number_of_lines_between() == 5

    assert (
        sorted_padded_csv_file.number_of_lines_between(stop=3)
        == sorted_padded_csv_file.number_of_lines_between(stop=3.5)
        == 1
    )

    assert (
        sorted_padded_csv_file.number_of_lines_between(stop=7)
        == sorted_padded_csv_file.number_of_lines_between(stop=7.5)
        == 2
    )

    assert (
        sorted_padded_csv_file.number_of_lines_between(start=7, stop=15)
        == sorted_padded_csv_file.number_of_lines_between(start=6.5, stop=15.5)
        == 3
    )

    assert (
        sorted_padded_csv_file.number_of_lines_between(start=15, stop=19)
        == sorted_padded_csv_file.number_of_lines_between(start=14.5, stop=42)
        == 2
    )


def test_sorted_padded_csv_file(padded_file_path):
    with sorted_padded_csv_file(
        padded_file_path,
        ("c", int),
        ["d", "b"],
    ) as spcf:
        assert (
            spcf[15:19]
            == spcf[14.5:42]  # type: ignore
            == list(spcf.get(start=15, stop=19))
            == list(spcf.get(start=14.5, stop=42))
            == [(15, [16, 14]), (19, [20, 18])]
        )


def test_splitted_sorted_padded_csv_file(splitted_padded_csv):
    with sorted_padded_csv_file(splitted_padded_csv, ("c", int), ["d", "b"]) as spcf:
        assert (
            spcf[6.5:]
            == spcf[7:]
            == [(7, [8, 6]), (11, [12, 10]), (15, [16, 14]), (19, [20, 18])]
        )

        assert spcf[:3] == spcf[:3.5] == [(3, [4, 2])]



from datetime import datetime
from multiprocessing import Process
from multiprocessing.connection import Connection
from pathlib import Path
from typing import Any, List, Optional, Tuple, Union, cast

from .csv import selector

MARGIN = 0.2


class BackgroundProcessor(Process):
    """Offers a way to compute - in a background process - all data needed by pyqtgraph
    from pyqtgraph visible range.

    The main (caller) process can communicate with the background process through a
    pipe.

    If several items a present in the pipe, then `BackgroundSelector` will process only
    the last one.

    Send `None` to the pipe to stop `BackgroundSelector`

    Usage:
    ------
    connection, background_connection = Pipe()

    background_processor = BackgroundProcessor(
        dir_path, ("a", int), ["b", "d"], background_connection
    )

    background_processor.start()

    connection.send((None, None, 100))
    connection.recv() == (
        [1, 5, 9, 13, 17],
        {
            "b": Selected.Y(mins=[2, 6, 10, 14, 18], maxs=[2, 6, 10, 14, 18]),
            "d": Selected.Y(mins=[4, 8, 12, 16, 20], maxs=[4, 8, 12, 16, 20]),
        }
    )

    connection.send(4.5, 13.5, 100) == (
        [5, 9, 13],
        {
            "b": Selected.Y(mins=[6, 10, 14], maxs=[6, 10, 14]),
            "d": Selected.Y(mins=[8, 12, 16], maxs=[8, 12, 16]),
        }
    )

    connection.send(None)
    """

    def __init__(
        self,
        dir_path: Path,
        x_and_type: Tuple[str, type],
        ys: List[str],
        connection: Connection,
    ) -> None:
        """Initializer

        dir_path  : Directory where all the files (sampled and non sampled) are located.
                    Non sampled path name's HAS to be `0.csv`

        x_and_type: Name and the type of X value
        ys        : Name of Ys types
        connection: One side of the pipe
        """
        super().__init__()
        self.__dir_path = dir_path
        self.__x_and_type = x_and_type
        self.__ys = ys
        self.__connection = connection

    def run(self) -> None:
        with selector(self.__dir_path, self.__x_and_type, self.__ys) as sel:
            _, x_type = self.__x_and_type

            while True:
                item: Optional[
                    Tuple[Optional[float], Optional[float], int]
                ] = self.__connection.recv()

                if item is None:
                    self.__connection.send(None)
                    return

                while self.__connection.poll():
                    item = self.__connection.recv()

                    if item is None:
                        self.__connection.send(None)
                        return

                visible_start_float, visible_stop_float, resolution = item

                assert not (visible_start_float is None) != (
                    visible_stop_float is None
                ), (
                    "`visible_start_float` and `visible_stop_float` must be both set "
                    "to None or set to a value which is not None"
                )

                if visible_start_float is None and visible_stop_float is None:
                    selected = sel[::resolution]
                elif visible_start_float is not None and visible_stop_float is not None:
                    visible_range = visible_stop_float - visible_start_float
                    visible_range_with_margin = MARGIN * visible_range

                    start_float = visible_start_float - visible_range_with_margin
                    stop_float = visible_stop_float + visible_range_with_margin

                    start, stop = (
                        (
                            datetime.fromtimestamp(start_float),
                            datetime.fromtimestamp(stop_float),
                        )
                        if x_type is not float
                        else (start_float, stop_float)
                    )

                    selected = sel[start:stop:resolution]  # type: ignore
                else:
                    raise ValueError(
                        "`visible_start_float` and `visible_stop_float` must be both "
                        "set to None or set to a value which is not None"
                    )

                first_x, *_ = selected.xs

                xs = (
                    [x.timestamp() for x in cast(List[datetime], selected.xs)]
                    if isinstance(first_x, datetime)
                    else cast(List[float], selected.xs)
                )

                self.__connection.send((xs, selected.name_to_y))



# from typing import List
# import pytest

# from ..splitted_padded_text_file import BlockAndItemFinder


# @pytest.fixture
# def blocks_len() -> List[int]:
#     blocks = [[3, 4, 5], [6], [7, 8, 9, 10], [11, 12]]

#     return [len(block) for block in blocks]


# def test_block_and_item_finder_index_offset_0(blocks_len):
#     block_and_item_finder = BlockAndItemFinder(blocks_len, 0)

#     assert (len(block_and_item_finder)) == 10

#     with pytest.raises(IndexError):
#         block_and_item_finder[-11]

#     assert block_and_item_finder[0] == block_and_item_finder[-10] == (0, 0)
#     assert block_and_item_finder[1] == block_and_item_finder[-9] == (0, 1)
#     assert block_and_item_finder[2] == block_and_item_finder[-8] == (0, 2)
#     assert block_and_item_finder[3] == block_and_item_finder[-7] == (1, 0)
#     assert block_and_item_finder[4] == block_and_item_finder[-6] == (2, 0)
#     assert block_and_item_finder[5] == block_and_item_finder[-5] == (2, 1)
#     assert block_and_item_finder[6] == block_and_item_finder[-4] == (2, 2)
#     assert block_and_item_finder[7] == block_and_item_finder[-3] == (2, 3)
#     assert block_and_item_finder[8] == block_and_item_finder[-2] == (3, 0)
#     assert block_and_item_finder[9] == block_and_item_finder[-1] == (3, 1)

#     with pytest.raises(IndexError):
#         block_and_item_finder[10]


# def test_block_and_item_finder_index_offset_1(blocks_len):
#     block_and_item_finder = BlockAndItemFinder(blocks_len, 1)

#     assert (len(block_and_item_finder)) == 9

#     with pytest.raises(IndexError):
#         block_and_item_finder[-10]

#     assert block_and_item_finder[0] == block_and_item_finder[-9] == (0, 1)
#     assert block_and_item_finder[1] == block_and_item_finder[-8] == (0, 2)
#     assert block_and_item_finder[2] == block_and_item_finder[-7] == (1, 0)
#     assert block_and_item_finder[3] == block_and_item_finder[-6] == (2, 0)
#     assert block_and_item_finder[4] == block_and_item_finder[-5] == (2, 1)
#     assert block_and_item_finder[5] == block_and_item_finder[-4] == (2, 2)
#     assert block_and_item_finder[6] == block_and_item_finder[-3] == (2, 3)
#     assert block_and_item_finder[7] == block_and_item_finder[-2] == (3, 0)
#     assert block_and_item_finder[8] == block_and_item_finder[-1] == (3, 1)

#     with pytest.raises(IndexError):
#         block_and_item_finder[9]


# def test_block_and_item_finder_slice_offset_0(blocks_len):
#     block_and_item_finder = BlockAndItemFinder(blocks_len, 0)

#     assert block_and_item_finder[0:0] == [None, None, None, None]
#     assert block_and_item_finder[:1] == [slice(1), None, None, None]
#     assert block_and_item_finder[:2] == [slice(2), None, None, None]
#     assert block_and_item_finder[:3] == [slice(None), None, None, None]
#     assert block_and_item_finder[:4] == [slice(None), slice(None), None, None]
#     assert block_and_item_finder[:5] == [slice(None), slice(None), slice(1), None]
#     assert block_and_item_finder[:6] == [slice(None), slice(None), slice(2), None]
#     assert block_and_item_finder[:7] == [slice(None), slice(None), slice(3), None]
#     assert block_and_item_finder[:8] == [slice(None), slice(None), slice(None), None]

#     assert block_and_item_finder[:9] == [
#         slice(None),
#         slice(None),
#         slice(None),
#         slice(1),
#     ]

#     assert (
#         block_and_item_finder[:10]
#         == block_and_item_finder[:11]
#         == block_and_item_finder[:]
#         == [slice(None), slice(None), slice(None), slice(None)]
#     )

#     assert block_and_item_finder[3:4] == [None, slice(None), None, None]
#     assert block_and_item_finder[1:6] == [slice(1, None), slice(None), slice(2), None]

#     assert block_and_item_finder[:] == [
#         slice(None),
#         slice(None),
#         slice(None),
#         slice(None),
#     ]

#     assert block_and_item_finder[-3:] == [None, None, slice(3, None), slice(None)]


# def test_block_and_item_finder_slice_offset_1(blocks_len):
#     block_and_item_finder = BlockAndItemFinder(blocks_len, 1)

#     assert block_and_item_finder[0:0] == [None, None, None, None]
#     assert block_and_item_finder[:1] == [slice(1, 2), None, None, None]
#     # TODO: Should be `slice(1, None)` instead of `slice(1, 3)`
#     assert block_and_item_finder[:2] == [slice(1, 3), None, None, None]
#     assert block_and_item_finder[:3] == [slice(1, None), slice(None), None, None]
#     assert block_and_item_finder[:4] == [slice(1, None), slice(None), slice(1), None]
#     assert block_and_item_finder[:5] == [slice(1, None), slice(None), slice(2), None]
#     assert block_and_item_finder[:6] == [slice(1, None), slice(None), slice(3), None]
#     assert block_and_item_finder[:7] == [
#         slice(1, None),
#         slice(None),
#         slice(None),
#         None,
#     ]

#     assert (
#         block_and_item_finder[:9]
#         == block_and_item_finder[:10]
#         == block_and_item_finder[:]
#         == [
#             slice(1, None),
#             slice(None),
#             slice(None),
#             slice(None),
#         ]
#     )

#     assert block_and_item_finder[-3:] == [None, None, slice(3, None), slice(None)]



import json
from csv import DictReader
from datetime import datetime
from multiprocessing import Pipe, cpu_count
from pathlib import Path
from threading import Thread
from typing import Dict, Iterator, List, Optional, Tuple

import yaml  # type: ignore
from click import Choice
from click.utils import echo
from pydantic import ValidationError
from pyqtgraph import (
    DateAxisItem,
    FillBetweenItem,
    GraphicsLayoutWidget,
    PlotCurveItem,
    PlotItem,
    mkQApp,
    setConfigOptions,
)
from pyqtgraph.graphicsItems.PlotCurveItem import PlotCurveItem
from PySide6.QtGui import QIcon
from typer import Argument, Exit, Option, Typer, colors, get_app_dir, prompt, secho

from .background_processor import BackgroundProcessor
from .csv import pad_and_sample, pseudo_hash
from .csv.selector import Selected
from .interfaces import COLOR_NAME_TO_HEXA, Configuration

setConfigOptions(background="#141830", foreground="#D1D4DC", antialias=True)

ICON_PATH = Path(__file__).parent / "assets" / "icon-256.png"

app = Typer()

APP_DIR = Path(get_app_dir("csv-plot"))
FILES_DIR = APP_DIR / "files"
FILES_DIR.mkdir(parents=True, exist_ok=True)
CONFIG_PATH = APP_DIR / "config.json"


def default_configuration_directory_callback(
    default_configuration_directory: Optional[Path],
):
    if default_configuration_directory:
        with CONFIG_PATH.open("w") as file_descriptor:
            json.dump(
                {
                    "default_configuration_directory": str(
                        default_configuration_directory
                    )
                },
                file_descriptor,
            )

        echo()

        secho(
            "😃 Default configuration directory set to: ",
            fg=colors.BRIGHT_GREEN,
            nl=False,
            bold=True,
        )

        secho(
            f"{default_configuration_directory} 😃",
            fg=colors.BRIGHT_GREEN,
        )

        echo()

        default_configuration_directory.mkdir(parents=True, exist_ok=True)

        raise Exit()


def get_configuration_files(configuration_files_dirs: List[Path]) -> List[Path]:
    base_files = [item for item in configuration_files_dirs if item.is_file()]
    directories = [item for item in configuration_files_dirs if item.is_dir()]

    extra_files = [
        extra_file
        for directory in directories
        for extra_file in directory.iterdir()
        if extra_file.suffix == ".yaml"
    ]

    return base_files + extra_files


def get_default_configuration_files() -> Iterator[Path]:
    if not CONFIG_PATH.exists():
        echo()
        secho("❌ ERROR: ", fg=colors.BRIGHT_RED, bold=True, nl=False)

        secho(
            "No configuration file provided and "
            "no default configuration directory defined ❌",
            fg=colors.BRIGHT_RED,
        )

        secho(
            "You can either use a configuration file or directory with `-c` or "
            "`--configuration-file-or-directory` option, or define default "
            "configuration directory with:",
            bold=True,
        )

        secho(
            "$ csv-plot --set-default-configuration-directory",
            bold=True,
            fg=colors.CYAN,
        )

        echo()
        raise Exit()

    with CONFIG_PATH.open("r") as file_descriptor:
        configuration = json.load(file_descriptor)
        return Path(configuration["default_configuration_directory"]).iterdir()


@app.command()
def main(
    csv_path: Path = Argument(
        ...,
        help="CSV file to plot. This file must contain a header.",
        exists=True,
        file_okay=True,
        dir_okay=False,
        resolve_path=True,
    ),
    configuration_files_dirs: Optional[List[Path]] = Option(
        None,
        "-c",
        "--configuration-files-or-directories",
        help=(
            "A list of configuration files or directories containing configuration "
            "files. You can specify only one configuration file, or one directory, or "
            "mix of files and directories... If a directory is specified, CSV Plot "
            "explores recursively all subdirectories to find configuration files. If "
            "several configuration files match the CSV file, then CSV Plot ask you "
            "which one you want to use."
        ),
        exists=True,
        file_okay=True,
        dir_okay=True,
        resolve_path=True,
        show_default=False,
    ),
    set_default_configuration_directory: Optional[Path] = Option(
        None,
        help=(
            "Set a default directory where CSV Plot will recursively look for "
            "configuration files. Once done, no need to specify configuration file any "
            "more. If a configuration file is specified at launch while a default "
            "directory is specified, then the configuration file will supersede the "
            "default directory."
        ),
        file_okay=False,
        dir_okay=True,
        resolve_path=True,
        is_eager=True,
        callback=default_configuration_directory_callback,
    ),
):
    """🌊 CSV Plot - Plot CSV files without headaches! 🏄

    CSV Plot is a tool to efficiently plot your CSV files.
    You define a YAML configuration file which specify how your CSV file should be
    plotted (layout, colors, legend, units, etc...) and CSV Plot handles the heavy work
    for you.

    CSV Plot does respect your computer memory. This means CSV Plot only loads into
    memory the portion of file which has to be plotted. CSV Plot is able to plot files
    which are bigger than your memory, and has been tested with file larger than 100GB.
    """

    # Get CSV file columns names corresponding to floats
    def is_castable_to_float(value: str) -> bool:
        try:
            float(value)
            return True
        except ValueError:
            return False

    with csv_path.open() as csv_file:
        reader = DictReader(csv_file)
        first_row = next(reader)

    columns = {name for name, value in first_row.items() if is_castable_to_float(value)}

    # Get configurations files
    configuration_files = (
        get_configuration_files(configuration_files_dirs)
        if configuration_files_dirs
        else get_default_configuration_files()
    )

    # Get configurations which could correspond to the CSV file
    configuration_file_to_configuration_dict_maybe_none = {
        configuration_file: yaml.load(
            configuration_file.open("r"), Loader=yaml.FullLoader
        )
        for configuration_file in configuration_files
    }

    # If a YAML file is empty, then it will be parsed as `None`. It has to be filtered
    configuration_file_to_dict = {
        configuration_file: configuration_dict
        for configuration_file, configuration_dict in configuration_file_to_configuration_dict_maybe_none.items()
        if configuration_dict is not None
    }

    try:
        configuration_file_to_configuration = {
            configuration_file: Configuration(**configuration_dict)
            for configuration_file, configuration_dict in configuration_file_to_dict.items()
        }
    except ValidationError as e:
        secho("ERROR:", fg=colors.BRIGHT_RED, bold=True)
        secho(str(e), fg=colors.BRIGHT_RED)
        raise Exit()

    matching_file_to_configuration = {
        configuration_file: configuration
        for configuration_file, configuration in configuration_file_to_configuration.items()
        if configuration.variables <= columns
    }

    if len(matching_file_to_configuration) == 0:
        secho(
            "❌ ERROR: ",
            fg=colors.BRIGHT_RED,
            bold=True,
            nl=False,
        )

        secho(
            "No configuration file matching with CSV columns found ❌",
            fg=colors.BRIGHT_RED,
        )

        raise Exit()
    elif len(matching_file_to_configuration) == 1:
        chosen_configuration, *_ = matching_file_to_configuration.values()
    else:
        matching_files_configurations = list(matching_file_to_configuration.items())
        matching_files, configurations = zip(*matching_files_configurations)

        secho(
            "Multiple configuration files correspond:",
            fg=colors.BRIGHT_GREEN,
            bold=True,
        )

        for index, matching_file in enumerate(matching_files):
            secho(
                f"{index} - {matching_file}",
                fg=colors.BRIGHT_YELLOW,
            )

        choice = prompt(
            "Choose which one you want to use",
            type=Choice([str(item) for item in range(len(matching_files))]),
            show_choices=False,
        )

        chosen_configuration = configurations[int(choice)]

    x = chosen_configuration.general.variable

    secho("Process CSV file... ", fg=colors.BRIGHT_GREEN, bold=True, nl=False)
    pad_and_sample(csv_path, FILES_DIR, x, cpu_count())
    secho("OK", fg=colors.BRIGHT_GREEN, bold=True)

    win = GraphicsLayoutWidget(show=True, title=f"🌊 CSV PLOT 🏄")
    win.showMaximized()

    first_plot: Optional[PlotItem] = None
    variable_to_low_high: Dict[str, Tuple[PlotCurveItem, PlotCurveItem]] = {}

    def get_plot(layout_item: Configuration.LayoutItem) -> PlotItem:
        plot: PlotItem = win.addPlot(
            row=layout_item.x - 1,
            col=layout_item.y - 1,
            title=layout_item.title,
            axisItems=(
                {"bottom": DateAxisItem()}
                if chosen_configuration.general.as_datetime
                else {}
            ),
        )
        plot.showGrid(x=True, y=True)

        plot.setLabel(
            "bottom",
            text=chosen_configuration.general.label,
            units=chosen_configuration.general.unit,
        )

        plot.setLabel(
            "left",
            text=layout_item.label,
            units=layout_item.unit,
        )

        return plot

    position_to_plot: Dict[Tuple[int, int], PlotItem] = (
        {
            (layout_item.x, layout_item.y): get_plot(layout_item)
            for layout_item in chosen_configuration.layout
        }
        if chosen_configuration.layout is not None
        else {}
    )

    for curve in chosen_configuration.curves:
        color = COLOR_NAME_TO_HEXA[curve.color]
        low = PlotCurveItem(pen=color)
        high = PlotCurveItem(pen=color)
        fill = FillBetweenItem(low, high, color)

        if not (curve.x, curve.y) in position_to_plot:
            position_to_plot[(curve.x, curve.y)] = get_plot(
                Configuration.LayoutItem(position=f"{curve.x}-{curve.y}")
            )

        plot = position_to_plot[(curve.x, curve.y)]

        plot.addItem(low)
        plot.addItem(high)
        plot.addItem(fill)

        variable_to_low_high[curve.variable] = low, high

    first_plot, *plots = position_to_plot.values()

    for plot in plots:
        plot.setXLink(first_plot)

    date_time_formats = chosen_configuration.general.date_time_formats

    def date_time_parser(x: str, date_time_formats: List[str]) -> datetime:
        for date_time_format in date_time_formats:
            try:
                return datetime.strptime(x, date_time_format)
            except ValueError:
                pass

        raise ValueError(
            f"time data '{x}' does not match any format in " "{date_time_formats}"
        )

    parser = (
        (lambda x: date_time_parser(x, date_time_formats))
        if date_time_formats is not None
        else float
    )

    connector, background_connector = Pipe()

    background_processor = BackgroundProcessor(
        FILES_DIR / pseudo_hash(csv_path, x),
        (x, parser),  # type: ignore
        list(chosen_configuration.variables),
        background_connector,
    )

    def on_sig_x_range_changed():
        x_range, _ = first_plot.viewRange()
        x_min, x_max = x_range

        connector.send((x_min, x_max, int(first_plot.width())))

    def update():
        while True:
            item: Optional[Tuple[List[float], Dict[str, Selected.Y]]] = connector.recv()

            if item is None:
                return

            xs, variable_to_y = item

            for variable, y in variable_to_y.items():
                low, high = variable_to_low_high[variable]
                low.setData(xs, y.mins)
                high.setData(xs, y.maxs)

    update_thread = Thread(target=update)
    update_thread.start()

    try:
        first_plot.sigXRangeChanged.connect(on_sig_x_range_changed)
        background_processor.start()
        connector.send((None, None, int(first_plot.width())))

        app = mkQApp()
        app.setWindowIcon(QIcon(str(ICON_PATH)))
        app.exec()
    finally:
        connector.send(None)



import hashlib
import os
from multiprocessing import Pool
from pathlib import Path
from typing import IO, Dict, List, Optional, Set, Tuple

from fast_pad_and_sample import pad as fast_pad
from fast_pad_and_sample import sample as fast_sample
from fast_pad_and_sample import sample_sampled as fast_sample_sampled


def pseudo_hash(path: Path, string: str = "") -> str:
    """Compute a pseudo hash based on :
    - The file size
    - The file last modification date
    - A given string

    (Real hash is too long to compute for big file.)
    """
    string = "-".join([str(os.path.getsize(path)), str(os.path.getmtime(path)), string])
    return str(hashlib.md5(bytes(string, "utf-8")).hexdigest())


def compute_chunks(file_path: Path, nb_chunks: int) -> List[Tuple[int, int]]:
    """Take a `file_descriptor` to a (non padded) text file, the file size and a
    number of chunks. Outputs a list of tuple.
    For each tuple, the first item represents the start byte index of a chunk, and the
    second item represents the end byte index of the same chunk.

    Example, with the following file:
    ---------------------------------

    a,b,c,d
    1,2,3,4
    5,6,7,8
    9,10,11,12
    13,14,15,16
    17,18,19,20

    compute_chunks(<path_file>, 4) == [
            (0, 16),
            (16, 35),
            (35, 47),
            (47, 59),
        ]
    """

    def get_next_new_line_byte_index(
        file_descriptor: IO, arithmetic_chunk_start: int
    ) -> int:
        file_descriptor.seek(arithmetic_chunk_start)
        return arithmetic_chunk_start + len(next(file_descriptor))

    arithmetic_chunk_size = file_path.stat().st_size / nb_chunks

    arithmetic_chunks_start = (
        int(index * arithmetic_chunk_size) - 1 for index in range(1, nb_chunks + 1)
    )

    with file_path.open() as file_descriptor:
        new_lines_byte_index_with_duplicate = [0] + [
            get_next_new_line_byte_index(file_descriptor, arithmetic_chunk_start)
            for arithmetic_chunk_start in arithmetic_chunks_start
        ]

    new_lines_byte_index = list(dict.fromkeys(new_lines_byte_index_with_duplicate))

    return list(zip(new_lines_byte_index[:-1], new_lines_byte_index[1:]))


def are_files_fully_sampled(dir_path: Path) -> bool:
    paths = sorted(dir_path.glob("*.csv"), key=lambda item: int(item.stem))

    for index, path in enumerate(paths):
        with path.open() as lines:
            if index == 0:
                next(lines)

            next(lines)

            try:
                next(lines)
                return False
            except StopIteration:
                pass

    return True


def pad(
    input_path: Path,
    output_path: Path,
    start_byte: Optional[int] = None,
    stop_byte: Optional[int] = None,
) -> None:
    """Pad the text file (in place) pointed by `input_path` with white spaces.

    input_path: The path of the input file
    output_path: The path of the output file. If not provided, will replace the input
                 file
    """
    output_path.parent.mkdir(parents=True, exist_ok=True)

    real_start_byte = 0 if start_byte is None else start_byte
    real_stop_byte = input_path.stat().st_size if stop_byte is None else stop_byte

    fast_pad(str(input_path), str(output_path), real_start_byte, real_stop_byte)


def sample(
    source_path: Path,
    dest_path: Path,
    x: str,
    period: int,
    start_byte: Optional[int] = None,
    stop_byte: Optional[int] = None,
) -> None:
    """Sample a CSV file every `period` line.

    Example:
    With the following CSV corresponding to `source_path`:
    x,a,b,c,d
    1,2,3,4,5
    2,3,0,7,6
    3,4,7,4,8
    4,8,6,4,3
    5,9,8,7,6
    6,5,7,8,2

    With x == "x"

    The CSV file written in `dest_path` will be:
    x,a_min,a_max,b_min,b_max,c_min,c_max,d_min,d_max
    1,2.0,3.0,0.0,3.0,4.0,7.0,5.0,6.0
    3,4.0,8.0,0.0,7.0,4.0,7.0,6.0,8.0
    5,5.0,9.0,7.0,8.0,7.0,8.0,2.0,6.0,
    """

    def has_to_be_excluded(value: str) -> bool:
        try:
            float(value)
            return False
        except:
            return True

    real_start_byte = 0 if start_byte is None else start_byte
    real_stop_byte = source_path.stat().st_size if stop_byte is None else stop_byte

    with source_path.open() as source_file, dest_path.open("w") as dest_file:
        not_stripped_header_line = next(source_file)
        header_line = not_stripped_header_line.rstrip()
        headers = header_line.split(",")

        x_index, *trash = [index for index, header in enumerate(headers) if header == x]

        assert len(trash) == 0, "Multiple `x` in headers"

        first_line = next(source_file)
        first_line_values = first_line.split(",")

        indexes = {
            index
            for index, value in enumerate(first_line_values)
            if not has_to_be_excluded(value)
        }.union({x_index})

        y_indexes = indexes - {x_index}
        y_headers = [headers[y_index] for y_index in y_indexes]

        if real_start_byte == 0:
            dest_headers = [x] + [
                item
                for sublist in [
                    [f"{header}_min", f"{header}_max"] for header in y_headers
                ]
                for item in sublist
            ]

            dest_file.write(f'{",".join(dest_headers)}\n')

    sorted_indexes = list(indexes)
    first_index, *_ = sorted_indexes

    deltas = [first_index] + [
        right - left for left, right in zip(sorted_indexes[:-1], sorted_indexes[1:])
    ]

    fast_sample(
        str(source_path),
        str(dest_path),
        sorted_indexes.index(x_index),
        deltas,
        len(deltas),
        period,
        len(not_stripped_header_line) if real_start_byte == 0 else real_start_byte,
        real_stop_byte,
    )


def sample_sampled(
    source_path: Path, dest_path: Path, period: int, has_header: bool
) -> None:
    """Sample an already sampled CSV file every `period` line.

    Example:
    With the following CSV corresponding to `source_path`:
    a,b_min,b_max,d_min,d_max
    1,2.0,6.0,4.0,8.0
    9,10.0,14.0,12.0,16.0
    17,18.0,18.0,20.0,20.0

    With period == 2,

    The CSV file written in `dest_path` will be:
    a,b_min,b_max,d_min,d_max
    1,2.0,14.0,4.0,16.0
    17,18.0,18.0,20.0,20.0
    """
    dest_path.parent.mkdir(parents=True, exist_ok=True)

    with source_path.open() as source_file, dest_path.open("w") as dest_file:
        line = next(source_file)
        str_values = line.split(",")
        _, *y_values = str_values
        nb_y_values = len(y_values)
        assert nb_y_values % 2 == 0

        if has_header:
            dest_file.write(line)

    fast_sample_sampled(
        str(source_path), str(dest_path), nb_y_values, period, has_header
    )


def pad_to_the_end(nb_workers: int, global_dir: Path, index: int) -> None:
    current_sampled_dir = global_dir / f"{index}_sampled"

    if not current_sampled_dir.exists():
        return

    padded_dir = global_dir / f"{index}"
    sampled_paths = list(current_sampled_dir.glob("*.csv"))
    padded_paths = [padded_dir / sampled_path.name for sampled_path in sampled_paths]
    arguments = list(zip(sampled_paths, padded_paths))

    with Pool(nb_workers) as pool:
        pool.starmap(pad, arguments)

    for sampled_path in sampled_paths:
        sampled_path.unlink()

    current_sampled_dir.rmdir()
    pad_to_the_end(nb_workers, global_dir, index + 1)


def sample_sampled_to_the_end(
    nb_workers: int, sampled_global_dir: Path, index: int
) -> None:
    current_sampled_dir = sampled_global_dir / f"{index}_sampled"

    if are_files_fully_sampled(current_sampled_dir):
        return

    next_sampled_dir = sampled_global_dir / f"{(index + 1)}_sampled"
    current_sampled_paths = list(current_sampled_dir.glob("*.csv"))

    next_sampled_paths = [
        next_sampled_dir / current_sampled_path.name
        for current_sampled_path in current_sampled_paths
    ]

    arguments = [
        (current_sampled_path, next_sampled_path, 2, index == 0)
        for index, (current_sampled_path, next_sampled_path) in enumerate(
            zip(current_sampled_paths, next_sampled_paths)
        )
    ]

    with Pool(nb_workers) as pool:
        pool.starmap(sample_sampled, arguments)

    sample_sampled_to_the_end(nb_workers, sampled_global_dir, index + 1)


def pad_and_sample(
    source_csv_file_path: Path, dest_dir_path: Path, x: str, nb_workers: int
) -> bool:
    """Pad and sample `source_csv_file_path` into `dest_dir_path` with `x`.

    If the file is already sampled, this function does not resample it but exits
    immediately without error.
    """

    dir_path = dest_dir_path / pseudo_hash(source_csv_file_path, x)

    try:
        dir_path.mkdir(parents=True)
    except FileExistsError:
        success_file = dir_path / "SUCCESS"

        if success_file.exists():
            return False

    padded_path = dir_path / "0"
    padded_path.mkdir(parents=True, exist_ok=True)

    sampled_path_1 = dir_path / "1_sampled"
    sampled_path_1.mkdir(parents=True, exist_ok=True)

    chunks = compute_chunks(source_csv_file_path, nb_workers)

    arguments_pad = [
        (source_csv_file_path, padded_path / f"{index}.csv", start_byte, stop_byte)
        for index, (start_byte, stop_byte) in enumerate(chunks)
    ]

    with Pool(nb_workers) as pool:
        pool.starmap(pad, arguments_pad)

    arguments_sample = [
        (
            source_csv_file_path,
            sampled_path_1 / f"{index}.csv",
            x,
            2,
            start_byte,
            stop_byte,
        )
        for index, (start_byte, stop_byte) in enumerate(chunks)
    ]

    with Pool(nb_workers) as pool:
        pool.starmap(sample, arguments_sample)

    sample_sampled_to_the_end(nb_workers, dir_path, 1)
    pad_to_the_end(nb_workers, dir_path, 1)

    success_file = dir_path / "SUCCESS"
    success_file.touch()

    return True
